'\"
'\" Copyright (c) 1994 Tom Moore, Natural Resources Canada
'\"               Petawawa National Forestry Institute
'\" All rights reserved.
'\"
'\" Permission is hereby granted, without written agreement and without
'\" license or royalty fees, to use, copy, modify, and distribute this
'\" documentation for any purpose, provided that the above copyright
'\" notice and the following two paragraphs appear in all copies.
'\"
'\" IN NO EVENT SHALL Natural Resources Canada BE LIABLE TO ANY PARTY
'\" FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
'\" ARISING OUT OF THE USE OF THIS DOCUMENTATION, EVEN IF Natural Resources
'\" Canada HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'\"
'\" Natural Resources Canada SPECIFICALLY DISCLAIMS ANY WARRANTIES,
'\" INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
'\" AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
'\" ON AN "AS IS" BASIS, AND THE Natural Resources Canada HAS NO OBLIGATION TO
'\" PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
'\" 
'\" SDB command created by Tom Moore and Carey Lockwood
'/" 
.so man.macros
.HS sdb cmds
.BS
'\" Note:  do not modify the .SH NAME line immediately below!
.SH NAME
sdb \- Open and access an SDB database
.SH SYNOPSIS
\fBsdb open\fI \fIcursorName file\fR
.PP
\fBsdb create\fI \fIcursorName nrecs\fR
.BE

.SH DESCRIPTION
.PP
The \fBsdb\fR command opens an SDB database file (given by the \fIfile\fR 
argument) and creates a database cursor (given by the
\fIcursorName\fR argument).  In the second form, the \fBsdb\fR command
creates an empty database cursor, which is preallocated to hold
\fInrecs\fR rows of records.
The \fBsdb\fR command returns its
\fIcursorName\fR argument.  At the time this command is invoked,
there must not exist a command or procedure named \fIcursorName\fR.
.PP
A cursor provides access to a database file.  When a cursor is created
it reads the entire database (identified by the argument \fIfile\fR) into
memory.  Cursor commands are provided to inspect and alter the database
definition and to retrieve and store values in the database.  
.PP
The cursor maintains a description of the state of the database, which
may be altered and examined.  Besides the values in the initial database,
cursor state includes a 'current record position', a selection mask, 
a sorted order, a presentation order for columns, as well as other
advanced database functions such as virtual fields,
joined tables, interpolated curves, production rules and deductive rules. 
.PP
The cursor also sets up a global array variable to that can be accessed
to obtain data elements from the table.  The global array variable is
indexed in two ways:  a one-dimensional array by data element name for 
the current record, and a two-dimensional array by row and column number.
Both arrays have the same name as the cursor (for example, if the 
\fIcursorName\fR is 'x', the array is named $x).
.PP
The one-dimensional array contains element that are referenced by the
names of the data elements.  For example, if \fIcursorName\fR 'x' 
contains a database element named 'SITE', the value of this element can be
accessed through the global value '$x(SITE)'.  This method retrieves
a value for that column from the 'current record'.  The current record
pointer can be manipulated to retrieve records from other rows in the
database.  Here is an example one-dimensional array access:
.DS
# Loop through the database and update some records in cursor 'x'
for {set pos 0} {$pos < [x info num_records]} {incr pos} {
  x position $pos
  if {$x(CODE) == "Po"} {
    puts "Changing species code on record $pos"
    set x(CODE) "Poplar"
  }
}
.DE

.PP
The two-dimensional array contains elements that are referenced by row
and column.  The first index in the array references row numbers, starting
at zero and increasing to (maximum row number - 1).  Row -1 is a 'virtual'
row that contains columns headings.  Data elements are stored in columns
starting at zero and increasing to the maximum column number - 1.  
Column numbers less than zero contain special 'pseudo-variables'
that are not actually part of the database.  At present the only 
pseudo-variable defined is in column -1:  this contains the record 
number.  The most useful purpose for two-dimensional access is to 
map the database into a spreadsheet table (using the tkTable widget).  
There is no whitespace in the two-dimensional index name:  the first
column of the first row would be retrieved as $x(1,1).

.SH "DATA TYPES"
The database supports two data types:  \fICODE\fR and \fINUMBER\fR.
\fICODE\fR variables represent arbitrary length strings or `labels'.
\fINUMBER\fR data types are stored in double precision floating point 
format and have a range of (+/-)10e(+/-)132 with 15 significant
digits of precision.  Boolean values are used in logical expressions, and
infered from \fINUMBER\fR data types with a value of zero representing
false and any other value representing true.

.SH "DATA FILE FORMATS"
The SDB extension uses a simple file format.  Data files are simple
column oriented ascii files.  There are two manditory record types:
a variable declaration record; and multiple data records.  The variable
declaration record lists the column names.  This must appear before
any data record in the file.  The record starts with "VAR" in column
1 and lists the names of the data columns seperated by whitespace.
CODE type variables are listed first, followed by a colon, and then NUMBER
data types are listed next.

.PP
Data records are in a similar format to the variable declaration
records.  The record starts with the keyword "DATA", and is followed 
by the data values seperated by white space.  CODE data values are listed 
in the same order as on the VAR statement, followed by a colon, and then
the NUMBER data values.

.PP
Data values containing whitespace or other special characters should 
be enclosed in braces ({}), following standard TCL treatment of list elements.

.PP
Here is a sample data file fragment:
.DS
    VAR STANDNUM SPECIES SITE : Age Area Height
    DATA 1 Pj 1 : 50 7 10.1
    DATA 2 Sb 1 : 35 15 8
    DATA 3 Pj 2 : 70 9 12.3
    DATA 4 {White Spruce} 1 : 35 12 7.6
    DATA 5 Sb 3 : 100 10 13.8
.DE
.PP
VAR statements may be also be interspersed in the data file, for example to 
indicate a change the column order.  It is thus possible to concatenate
several datafiles together, each of which has the same variables 
but not necessarily in the same order.

.PP
All records starting with other than "VAR" and "DATA" are ignored.

.PP
\fBJoin\fR files are a variant of the simple data file.  These are
files which may be joined in a relational fashion to the primary table
through the use of the \fBjoin\fR subcommand.  When join tables are
loaded into a database cursor matching columns are identified and an index is
built for fast location of matching records.  Columns not in common are
treated as additional columns in the database.  When a row in the database 
is referenced the matching join record is located and additional column
data is retrieved automatically.

.PP
The structure of a join file is identical to the simple data file format
with respect to VAR and DATA statements.  The interpretation of
the CODE data values in matching columns is slightly different.  
A "*" character in a matching column
will be interpreted as a wild card character.  Any record which fails
to find an exact match with a join file record can match to a join
file records containing wild card characters.  If a match cannot be
found and there are no wild card records, the added columns will contain
"." for CODE variables and 0 for numeric variables.

.PP
Matching columns in a join file may also contain a special expansion
operator: "/".  Fields which contain an expansion are split into component
parts, and the record is replicated with each part of the expansion.
For example, a field containing an expansion of "Sb/Sw/Pj"  will be
expanded into three records:  one each for Sb, Sw and Pj.  If multiple
fields contain expansions, the number of records added to the join table
is the cross product of each expansion field.  
.PP
Here is an example of a join file containing expansions:

.DS
    VAR SPECIES SITE Quality : HSI
    DATA Sb/Sw 1 Good : 1.0
    DATA Sb 2 Fair : 0.8
    DATA Sw 2 Poor : 0.4
    DATA Sb/Sw 3/4 Poor : 0.2
    DATA Pj * Medium : 0.5
    DATA * * {Not Suitable} : 0.0
.DE

.PP
If this join file was combined with the previous data file, the columns
SPECIES and SITE would be matching columns, and Quality and HSI would
be additional columns.
The first data record is expanded into two records:  one each for species
Sb and Sw.  The fourth data record is expanded into four records having 
Species-Site combinations of Sb 3, Sb 4, Sw 3, and Sw 4.  The fifth
record contains a wild cards character in the SITE column.  All records
having a SPECIES of Pj would match to this record.  The final data record
contains wildcards in both matching columns.  This join record will match 
to any database record that is not explicitly matched by the previous
join records.

.PP
Yield curve files are similar to join files.  A yield curve file contains
a number of CODE fields which are used to match yield curves to records
in the same manner as with join files.  Yield curve files contain numeric
columns which describe the vertices of curves plotting the relationship
between independent and dependent variables.  There are no predefined
names which must be used as column headings in yield curve files.  The
identification of independent and dependent variables occurs in the
\fBinterpolate\fR function, which is described later.

.PP
Here is an example of a yield curve file:
.DS
    VAR SPECIES SITE : Age Volume
    DATA Pj/Pw 1 : 0 0
    DATA Pj/Pw 1 : 30 0
    DATA Pj/Pw 1 : 70 200
    DATA Pj/Pw 1 : 90 250
    DATA Pj/Pw 1 : 110 0
    DATA Pj/Pw 2 : 0 0
    DATA Pj/Pw 2 : 35 0
    DATA Pj/Pw 2 : 75 170
    DATA Pj/Pw 2 : 100 210
    DATA Pj/Pw 2 : 130 0
    DATA Pj/Pw * : 0 0
    DATA Pj/Pw * : 40 0
    DATA Pj/Pw * : 80 150
    DATA Pj/Pw * : 110 180
    DATA Pj/Pw * : 150 0
.DE
.PP
The above yield curve file contains an expansion on the SPECIES column
to provide yield curves that match to Pj and Pw.  The SITE column explicitly matches the "1" and "2" site classes, plus a wildcard class that matches any
other Pj or Pw site class.

.PP
The records in a yield curve file must record the vertices of the relationship
of independent and dependent variables in monotonic increasing order of the
independent variable.

.SH "EXPRESSIONS"
.PP
\fBsdb\fR database cursors use expressions to compute values in
virtual fields, production rules, deductive rules, and selection
criteria.  The \fBsdb\fR command contains a builtin compiler to
parse, optimize and evaluate expressions according to a general
purpose language patterned loosely after 'C'.

The BNF grammar for the expressive language is summarized as follows:
.DS
expr             := <numeric_expr> | <code_expr>
                 := (expr)
                 := assignment
                 := (expr; expr; ....)

assignment       := var = expr

numeric_expr     := numeric constant
                 := numeric database element
                 := numeric_expr <binary_op> numeric_expr
                 := <unary_op> numeric_expr
                 := <simplex_function>(numeric_expr)
                 := <duplex_function>(numeric_expr)
                 := <char_function>(code_expr)
                 := code_expr relational_op code_expr
                 := IN(code_expr, code_list)
                 := interpolate(setName, independent, dependent, expr)
                 := numeric_expr ? (relational_op, numeric_expr, numeric_expr):...
                 := code_expr ? (relational_op, code_expr, numeric_expr):...
                 := random(string)

binary_op        := <arithmetic_op> | <logical_op> | <relational_op>

arithmetic_op    := + - * / ^ %

logical_op       := && ||

relational_op    := == < <= > >= !=

unary_op         := - ! +

simplex_function := abs acos asin atan ceil cos cosh exp floor int log log10
                    round sin sinh sqrt tan tanh


duplex_function  := atan2 fmod hypot pow

code_expr        := string constant
                 := code database element
                 := numeric_expr ? (relational_op, numeric_expr, code_expr):...
                 := code_expr ? (relational_op, code_expr, code_expr):...
                 := code_function

code_list        := "code/code/code..."
                 := list_function

code_function    := none defined
list_function    := none defined
.DE

In general, there is no limit to the complexity in which the above
elements are combined to form expressions.

Although most of the syntax is like 'C', a few of the special
syntatic forms require a bit more explanation.

.TP
\fBinterpolate( \fIsetName\fB, \fIindependent\fB, \fIdependent\fB, \fIexpr\fB)\fR
The interpolate function computes a linear interpolation of a dependent value 
from a series of vertices describing a curve.  The function
takes four arguments to define the yield curve set to be used, the independent
and dependent axis, and the independent value.  The first argument
\fIsetName\fR is a character constant that refers to the yield curve
set loaded or created with the \fBcurve\fR subcommand (see below).
The second and third arguments are character constants that identify
the numeric columns from the yield curve file that will be used for the
independent and dependent (x and y) variables.  The fourth argument is a 
numeric expression that is evaluated to produce the independent value used in
the interpolation.  When evaluated, the \fBinterpolate\fR function scans 
the yield curve set for an entry that matches in a relation fashion to the 
current record.  The dependent and independent series of vertices are 
extracted, and the function then interpolates the the dependent 
value corresponding to the independent value.  If a matching entry cannot
be found in the yield curve set, the function returns zero.

.TP
\fInumeric_expr ? (relational_op, numeric_expr, numeric_expr):...:numeric_expr
.TP
code_expr ? (relational_op, code_expr, numeric_expr):...:numeric_expr
.TP
numeric_expr ? (relational_op, numeric_expr, code_expr):...:code_expr
.TP
code_expr ? (relational_op, code_expr, code_expr):...:code_expr\fR
The above four forms are variants of a 'choice' operator.  These forms
are similar to a 'if elseif ... else' construct.  The difference in the
above forms is that the first two return a numeric value, and the last two 
return a CODE value.  The first and third test a numeric expression, the
second and fourth test a CODE expression.  The first element of the form
is an expresssion of arbitrary complexity.  The expression is evaluated and
tested against the first clause according to the \fIrelational_op\fR.  If
the test evaluates to true (non-zero), the third expression in the clause
is evaluated and returned as the result.  If the test evaluates to false
(zero), then next clause is checked in the same manner and so on.  If 
none of the clauses evaluate to true, the final expression is evaluated
and returned as the default value.

.TP
\fBrandom( \fIstring\fB )\fR
The \fBrandom\fR functions is evaluated to return a random number 
from the stream defined as \fIstring\fR, where \fIstring\fR is a 
character constant.  See the \fBrandom\fR subcommand below for information
about defining random number streams.

.TP
\fBIN( \fIcode_expr\fB , \fIcode_list\fB )\fR
The \fBIN\fR function returns a one if the value of the evaluation of
\fIcode_expr\fR is found in the list specified by \fIcode_list\fR.  If
it is not found in the list, the function returns zero.

.SH "CURSOR COMMAND"
.PP
The \fBsdb\fR command creates a new Tcl command whose
name is \fIcursorName\fR.  The new
command may be used to invoke various
operations on the cursor.  It has the following general form:
.DS C
\fIcursorName option \fR?\fIarg arg ...\fR?
.DE
\fIOption\fR and the \fIarg\fRs
determine the exact behavior of the command.  The following
commands are possible for \fBsdb\fR cursors:

.TP
\fIcursorName \fBallocate\fR \fIfieldName type\fR
Create a new field within the database, and allocate storage so that
unique, persistant values can be stored in each row.  The name of the new 
field is specified by the \fIfieldName\fR argument.  The field type 
(specified by the \fItype\fR argument) may be 
either \fICODE\fR (arbitrary length alpha-numeric label), or
\fINUMBER\fR (stored in double precision floating point).  All rows in
the new field will be initialized to \fImissing\fR (in the case of
\fICODE\fR fields) or zero (in the case of \fINUMBER\fR fields).

.TP
\fIcursorName \fBarrange\fR \fIplacement destination source\fR
Arrange the presenation of fields within the database.  \fIplacement\fR
may be either \fBbefore\fR or \fBafter\fR.  The \fIsource\fR fields
are rearranged to display before or after the \fIdestination\fR field.

.TP
\fIcursorName \fBcalculate\fR \fIexpression\fR
Iterates through every selected record computing the value of \fIexpression\fR.
As a side effect may update fields in the database if the expression contains
an assignment.  The expression is
specified in a standard algebraic form such as `a=10*b' (see the previous 
section on expressions).   

.TP
\fIcursorName \fBclose\fR 
This command is invoked to destroy the cursor and free up memory
occupied by the database.  Changes made to the current database are not
automatically saved back into the original file.

.TP
\fIcursorName\fR \fBcurve\fR \fIoption\fR
Commands to load and manipulate curve sets.  Curves sets are similar to 
lookup tables: they may contain multiple class variables which are used to 
match the curves to database records.  
Labels are used to identify sets, and
all curves in a set must have identical class and numeric values. 
Once curves have been loaded into sets, the values may be accessed using 
the \fIinterpolate\fR function in numeric expressions.  The following 
options may be used with the curve command:
.RS
.TP
\fBcreate\fR \fIsetName {classNames} {valueNames}\fR
Creates a new curve set and identifies the class and value columns
that are common omong curves in this set.  Returns an error if
\fIsetName\fR already exists.
.TP
\fBdefine\fR \fIsetname {class_list} {value_list}\fR
the \fBdefine\fR option adds or replaces a curve within a set.
The set must already exist.  The number of items in \fI{class_list}\fR and
\fI{value_list}\fR must match the items defined for this set.
.TP
\fBfields\fR \fIsetName\fR
List the names of the fields in the curve set.  Two lists are returned:
the class names and the value names.
.TP
\fBget\fR \fIsetName {class_list}\fR
Retrieves a curve from the named set.  \fI{class_list}\fR provides
the combination of items that identifies the curve to retrieve.
The returned value is a list of lists, one for each numeric variable.
The lists are all equal in length.
.TP
\fBlist\fR \fIsetName\fR
Returns a list of names of the defined curves in the specified \fIset\fR.
.TP
\fBload\fR \fIsetName fileName\fR
Loads curves from a file in SDB format.  Creates the yield set if
it does not already exist.  If the set already exists the item names in 
the file must exactly match columns defined in the curve set.
.TP
\fBsets\fR
List all of the sets that have been defined so far.
.RE

.TP
\fIcursorName \fBdefine\fR \fIfieldName\fR \fIexpression\fR
Create a new virtual field within the database, based on the
expression specified in argument \fIexpression\fR.  The expression is
specified in a standard algebraic form such as `10*b'.   
The field type 
(either \fICODE\fR or \fINUMBER\fR) is automatically determined 
by the type of the expression.  This command compiles and stores
the expression for later evaluation.  The expression will be 
dynamically evaluated whenever a value is retreived for a virtual
field.  The virtual field will be defined as a type \fICONSTANT\fR if
the expression resolves as a constant, otherwise as a type 
\fIEXPRESSION\fR.  Upon success, \fIfieldName\fR is returned.
Returns an error if \fIfieldName\fR already exists or if the \fIexpression\fR
cannot be parsed correctly..

.TP
\fIcursorName \fBdelete\fR \fIfieldName\fR
Delete a field within the database specified by the argument
\fIfieldName\fR.  All storage used by the field is released.
Other virtual fields that use the deleted field in expressions
will be temporarily redefined to either \fImissing\fR or zero.

.TP
\fIcursorName \fBformat\fR \fIfieldName\fR \fIspec\fR
Provide a format specification string to display values in spreadsheet
cells or with the \fBget\fR command.  Format specifications are the
same as for the 'C' language printf routine.  The default specification
is "%s" for CODE values and "%-.15g" for NUMBER values.

.TP
\fIcursorName \fBget\fR \fIvalue\fR
Retrieves a row from the database.  \fIValue\fR specifies the row
number to retrieve, and may be an integer row number or the keyword
\fIcurrent\fR.  This command will return an error if \fIvalue\fR is out of 
the range of the database.

.TP
\fIcursorName \fBindex\fR \fIindexColumn\fR
The database will be indexed 
for keyed access based on the specified column.  \fBseek\fR commands
will find records matching the \fIindexColumn\fR.   If a database
field named \fIindexColumn\fR does not exist an error will be returned.
Databases can not be indexed on columns defined as expressions or on
lookup table elements.

.TP
\fIcursorName \fBinfo\fR \fIoption\fR
Returns information about the current state of the cursor.
\fIOptions\fR may be:
.RS
.TP
\fBcodes\fR \fIfieldName\fR
Returns a list of unique code values related to the database field
identified by argument \fIfieldName\fR.  An error will be returned 
if \fIfieldName\fR is not of type \fICODE\fR.
.TP
\fBcurrent\fR
Returns the current record number pointed to by the file pointer.
.TP 
\fBeof\fR
Returns `1' if the file pointer is currently positioned at the
last record, or returns `0' otherwise.
.TP
\fBitem\fR \fIfieldName\fR
Returns a description of the database field named \fIfieldName\fR.
The first item in the field descrition is either \fIALLOCATED\fR,
\fICONSTANT\fR, or \fIEXPRESSION\fR.  The second item in the
field descrition is either \fICODE\fR or \fINUMBER\fR.  If the
field type is \fICONSTANT\fR, the third item is the value of the
field.  If the field type is \fIEXPRESSION\fR, the next two values
are the expression in it's original and optimized forms.
.TP
\fBitems\fR
Returns a list of the fieldNames in this database.
.TP
\fBname\fR
Returns the file name of the database.
.TP
\fBnum_items\fR
Returns a count of the number of fields currently defined in 
the database.
.TP
\fBnum_output\fR
Returns a count of the number of fields automatically output,
such as by the \fBget\fR command.
.TP
\fBnum_records\fR
Returns a count of the number of records in in the database.
.TP
\fBnum_selected\fR
Returns a count of the number of currently selected records.
.TP
\fBrange\fR \fIfiledName\fR
Returns the minimum and maximum value of the variable.  An
error will be returned if \fIfieldName\fR is not a numeric 
variable.
.RE

.TP
\fIcursorName \fBjoin\fR \fIoption\fR \fI?args?\fR
The \fBjoin\fR command manages tables which are joined in a relational
fashion to the main table.  \fIOptions\fR may be:

.RS
.TP
\fBload\fR \fItableName\fR \fIfileName\fR
Specifies the name of a file that will be loaded in as a relational
table.  The file will be loaded into a relational table that is given a
symbolic name of \fItableName\fR.  The file must exist in the simple 
database format.  If
the relational table does not already exist it will be created
and the specified file will be loaded into it.  If the relational
table already exists, the current rows will be deleted and new rows
will be loaded from the specified file.

The columns upon which the join is based are determined automatically
by matching names from the existing database to the columns in 
\fIfileName\fR.  Only 'CODE' type columns may be used to create common
elements upon which the join will be based.  'NUMERIC' data elements 
may exist in \fIfileName\fR.  These will 
exist in the projection of the join, but may not be used as 'matching'
columns.  If you need to use a numeric data element as matching column 
in the join, define and expression that converts the value to a string
first.

\fIfileName\fR does not have to be sorted in any 
particular order:  when the table is joined a balanced binary tree index 
is created
to facilitate rapid access.  Once \fItableName\fR
has been created the additional columns in the new table become
part of the database and may be used anywhere that any other
type of variable may be used.
.TP
\fBsave\fR \fItableName\fR \fIfileName\fR
Saves the contents of a relational table into the specified
\fIfileName\fR.  The file will be saved in the simple database format.

.TP
\fBinfo\fR \fItableName\fR \fIoption\fR
Returns information about the relational table.  Valid options
to the \fBinfo\fR command are \fBindex\fR, \fBcodes\fR, and \fBvalues\fR.
\fBinfo index\fR will return a list of the variable names that are in common 
and specify the relation
between the base table and the joined table.  \fBinfo codes\fR returns
a list (possibly empty) of the variable names that are the additional
code (label) fields in the joined table.  \fBinfo values\fR returns a
list (possibly empty) of the variable names that are the additional 
value (numeric) fields in the joined table.

.TP
\fBdelete\fR \fItableName\fR \fIindex\fR
Deletes a row from the specified \fItableName\fR.  The \fIindex\fR
paramter is a list of values that identifies the row to be deleted.

.TP
\fBinsert\fR \fItableName\fR  \fIindex codes values\fR
Inserts a row into the specified \fItableName\fR.  The \fIindex\fR
paramter is a list of values that identifies the row to be inserted.
The \fIcodes\fR and \fIvalues\fR parameters are lists that specify 
the codes and values to be inserted into the row specified by the \fIindex\fR
parameter.  If the row already exists in the table, it is replaced.

.TP
\fBconnect\fR \fItableName\fR \fIremoteTable\fR
The \fBconnect\fR command established a relation between the current
database and a relational table defined in another database.  In this
way relational tables may be shared between active databases.  
Shared tables may reduce memory consuption and improve performance.

The \fIremote\fR relational table must exist in a currently active
database.  The remote file specification is \fItableName\fR@\fIdatabase\fR,
where \fIdatabase\fR is the name of the remote database cursor, and
\fItableName\fR is the name of the existing joined table within that
database.

Connected tables behave in the same manner as regular relational
tables.  Values may be changed, records may be inserted and deleted.
Since only a single copy of the relational
table exists, changes to the records in the table are immediatley visible to 
all connected databases.

Although tables may be shared, memory for the relational table is
allocated in the database which originally loaded the table.  When the
parent database is closed, or if the relational table is deleted from
the parent database, the records in the shared relational table will
be deleted.  Although connecting databases will have defined variable
names for column in the shared table, no records will exist, and insertions
will not be allowed.
.RE
 
.TP
\fIcursorName \fBoutput\fR ?\fIoptions\fR? ?\fIvalue...\fR?
Define the fields that will be automatically output using commands
such as \fBget\fR.  \fIOptions\fR may be `on' or `off'.  \fIValues\fR
list the names of the database fields that are to be toggled on or
off.  If the \fBoutput\fR command is issued without any options or
values it will return a list of the fields that are currently toggled
on for output.

Note that columns in the database must be flagged for output
if they are to accessible in the two-dimensional array.  

.TP
\fIcursorName \fBposition\fR \fIvalue\fR
Change the current value of the file pointer.  \fIValue\fR must be a
valid record number (within range of the size of the file).  Note that
the first record in the database is at row '0'.

.TP
\fIcursorName \fBput\fR \fIrecord\fR \fIlist\fR
Updates the database record \fIrecord\fR with the values specified 
in \fIlist\fR.

.TP
\fIcursorName \fBrandom\fR \fBsequenceName\fR \fIoption\fR
Defines and modifies random number sequences, each of which is given 
a \fBsequenceName\fR.  Random sequences are used by the \fBrandom\fR
function.  Sequences do not need to be explicitly created, they are
automatically generated when referenced.  Automatically generated
sequences are seeded with a value of '1'.  The 'C' library has the
following to say about random number sequences:
.DS
     random() uses a non-linear additive feedback  random  number
     generator employing a default table of size 31 long integers
     to return successive pseudo-random numbers in the range from
     0  to (2**31)-1.  The period of this random number generator
     is very large, approximately 16*((2**31)-1).
.DE
In the Sdb implementation, the random values are converted to floating
point and scaled according to the requested distribution (see below).  
The following \fIoptions\fR may be used:
.RS
.TP
\fBseed\ \fIvalue\fR
Redefines the seed to the random number generator for the named
sequence.  If the sequence does not exist, it is created.
.TP
\fBtype\ \fIdistribution\fR
Specifies the type of random distribution for the named sequence.  
\fIDistribution\fR may be
either "uniform" or "normal".  Values generated by the "uniform" distribution
are scaled between 0.0 and 1.0.  Values generated by the normal distribution
have a mean of 0.0 and a standard deviation of 1.0.
If the sequence does not exist, it is created.
.RE

.TP
\fIcursorName \fBreload\fR \fIfile\fR
Deletes all allocated fields within the database and creates new
allocated fields which are loaded from \fIfile\fR.  The number of records 
in the database may change as a result of the \fBreload\fR command.
Constants, expression, yield curves and relational tables are unchanged
as a result of the \fBreload\fR command.

.TP
\fIcursorName \fBseek\fR \fIkey\fR
Change the current value of the file pointer by finding a matching
key in an indexed column.  If the \fIkey\fR cannot be matched in the
indexed column an error is returned and the record pointer is not
changed.  To use this command the indexed column must have been
identified when the database was opened (see the \fBsdb\fR 
\fIopen\fR command).


.TP
\fIcursorName \fBselect\fR ?\fIoption\fR?
The \fBselect\fR command manages a selection mask that is used to filter 
records.  Many of the other commands that operate on cursors apply the
selection mask to filter records for processing.  For example, the
\fBcalculate\fI command will only apply its expression to slected records.
Some of the \fBselect\fR sub-commands require an expression.  Expressions
are evaluated in the same manner as 'C' where a value of zero represents
a logical false and anything else represents true.
The following options are available to manipulate and query the selection mask:
.RS
.TP
\fBand\fR ?\fIexpression\fR?
\fBand record\fR ?\fIrec\fR?
Performs a logical `and' on the selected set.  Reduces the selected set 
to include only those records that are already selected and also
match the criteria specified in the expression.  If no expression is
provided the command produces an empty selected set (no records selected).

.TP
\fBfirst\fR
Returns the number of the first record in the selected set.  
If there are no selected records this returns an error.

.TP
\fBlast\fR
Returns the number of the last record in the selected set.
If there are no selected records this returns an error.

.TP
\fBnext\fR
Returns the number of the next record in the selected set beyond
the current record pointer.
If there are no selected records, or if the current record pointer is
beyond the last selected record this returns an error.

.TP
\fBnot\fR ?\fIexpression\fR?
Performs a logical `not' on the selected set.  If an expression is
supplied the command will toggle the selection status of all records
in the database that match the expression.  If no expression is supplied,
the selection mask is inverted.

.TP
\fBpop\fR
The current selection mask is discarded and the select mask on the 
top of the stack is made current.  If the stack is empty this command
returns an error.
.TP
\fBpush\fR
A copy of the current selection mask is pushed on to the top of a stack.
the \fBpush\fR and \fBpop\fR commands provide a convenient means to
temporarily store and retrieve the contents of the selection mask.
.TP
\fBset\fR ?\fIexpression\fR?
Performs a logical `or' on the selected set.  Increases the selected 
set to include all records in the database that match the value of the
expression.  If an expression is not supplied then all records in the
database become selected.
.TP
\fBstatus\fR \fIrecord\fR
Tests the status of \fIrecord\fR.  If selected, returns a "1".  If not
selected, returns a "0".
.TP
\fBunset\fR ?\fIexpression\fR?
Performs a logical `clear' on the selected set.  Decreases the selected 
set to exclude all records in the database that match the value of the
expression.  Without an expression clears all records.

.RE

Here is an example of the use of the select command:
.DS
# skip through some record in a cursor 'x'

# first, select all records
x select set

# next, limit the selection
x select and {CODE=="Po" && SITE == "1" && Maturity == "regen"}

# now, loop through the list
x position [x select first]
while {1} {
  puts "Found a match in stand number $x(STANDNUM)"
  if {[x info current] == [x select last]} break
  x position [x select next]
}
.DE

.TP
\fIcursorName\fR \fBsnapshot\fR \fIfileName\fR
Dumps the selected records out to the named file, in SDB records format.
Files created in this way can be opened as inventory files (\fBsdb open\fR),
lookup tables (\fIcursorName\fR \fBjoin load\fR), or as curve sets
(\fIcursorName\fR \fBcurve load\fR).  If the file already exists it
is overwritten.

.TP
\fIcursorName\fR \fBsort\fR \fIfield\fR...
Sort the database in order according to the specified fields.  There
is no limit to the number of fields that may be used to specify the 
sorting order.  Each field may be followed by an optional \fIorder\fR
prefix ("+" or "-") to modify the collation sequence.  The default order 
if not specified is 'ascending'.


.SH KEYWORDS
sdb, cursor, database

.SH BUGS
Possibly, although I have tested extensively using purify.

.SH AUTHOR
Tom Moore, tmoore@pnfi.forestry.ca