/* -*-C-*-  */
%{

/*
 * This is a FLEX input file for parsing CALCULATE and DEFINE expressions.
 *
 * Copyright (c) 1995 Tom Moore.  All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all copies of this software.
 * 
 */


/*------------------------------------------------------------
 *
 * Macro definition for lex input function
 *
 *----------------------------------------------------------*/
#define YY_INPUT(buf,result,max_size) \
 result = ((sdbTextPos < sdbTextLen) ? (*buf = sdbTextBuff[sdbTextPos++],1) : ((sdbTextPos==sdbTextLen)?(*buf = 2,sdbTextPos++,1):0))

/*
 * If not using flex, then probably (hopefully!) using the ATT lexer.
 * This doesn't define a YY_INPUT macro, but instead gets input
 * from getc().  Redefine gets() as a macro to provide the input
 * stream from the interal buffer.  I hope that this code is constant between
 * various flavours of lex!  
 */
#ifndef FLEX_SCANNER
#define getc(infile) ((sdbTextPos < sdbTextLen) ? sdbTextBuff[sdbTextPos++] : (sdbTextPos==sdbTextLen)?(sdbTextPos++,2):0)
#endif

static void          Sdberror _ANSI_ARGS_((char* s));
static void          analyze_buff_init _ANSI_ARGS_((char CONST *string));

/*
 *------------------------------------------------------------
 *
 * Sdberror --
 *
 *      Called by the parser when an error occurs.  Trys to
 *      figure out what went wrong an generate an informative error
 *      message.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      May add several lines to the result string.
 *
 *------------------------------------------------------------
 */
static void 
Sdberror(s)
  char* s;
{
  CONST char *token;
  char *temp,buffer[88];
  int first,length,space,i;

  /*
   * Figure out where in the buffer the error occured,
   * and print this out.
   */
  first =(sdbTextPos>30)?sdbTextPos-30:0;
  length = (first+60)>sdbTextLen?sdbTextLen:60;
  temp = (char*)ckalloc((unsigned int)length+1);
  strncpy(temp,sdbTextBuff+first,(unsigned int)length);
  temp[length]=0;

  space=27+(sdbTextPos-first-strlen(Sdbtext));
  Tcl_AppendResult(CurrentInterp,"A parse error occured near \"",temp,"\"\n",NULL);
  ckfree(temp);
  for(i=0;i<space;i++) buffer[i]=' ';
  buffer[space]='^';
  for(i=0;i<strlen(Sdbtext);i++) buffer[space+i]='^';
  buffer[space+(strlen(Sdbtext)?strlen(Sdbtext):1)]=0;
  Tcl_AppendResult(CurrentInterp,buffer,"\n",NULL);

#ifdef USING_BISON
  /*
   * Look at the current token.  This may give
   * a clue as to what went wrong.
   *
   * The first case is illegal characters in the input stream.    */
  token=yytname[YYTRANSLATE(Sdbchar)];
  if (strcmp("$illegal.",token)==0) {
    Tcl_AppendResult(CurrentInterp,"while parsing \"",Sdbtext,
		     "\", which is not a valid language element.\n",
		     "Possible cause:  bad characters in input, or missing/extra quotes in string",NULL);
    return;
  }

  /*
   * Unknown symbol
   */
  else if (strcmp("UNKNOWN",token)==0) {
    Tcl_AppendResult(CurrentInterp,"while parsing \"",Sdbtext,
		     "\", which is an unknown variable.\n",
		     "Possible cause:  spelling mistake.",NULL);
    return;
  }

  /*
   * Premature end of input
   */
  else if (strcmp("EOI",token)==0) {
    Tcl_AppendResult(CurrentInterp,"at the end of input\n",NULL);
  }

  /*
   * End of statement marker
   */
  else if (strcmp("EOS",token)==0) {
    Tcl_AppendResult(CurrentInterp,"at the end of a statement\n",NULL);
  }

  /*
   * Anything else, just state what was being parsed
   */
  else {
    Tcl_AppendResult(CurrentInterp,"while parsing \"",Sdbtext,"\"",NULL);
    if (*token && strncmp(token+1,Sdbtext,strlen(token)-2))
      Tcl_AppendResult(CurrentInterp," which is a ",token," token.",
		       NULL);
  }

  /*
   * Now take a peek at the error string generated by yyparse.
   * If it was an automatically generated BISON error, just pull out
   * and reformat the interesting bits.
   */
  if (strncmp(s,"parse error",11)==0) {
    if (strcmp("parser error, expecting `EOS'",s)==0) {
      Tcl_AppendResult(CurrentInterp,
	 "In this context the parser was expecting the end of the statement\n",
	 "Possible cause:  mismatched parenthesis",NULL);
      return;
    }
    Tcl_AppendResult(CurrentInterp,
	   "\nIn this context the parser was",s+12,NULL);

    /*
     * Type mismatc
     */
    if (strcmp(token,"STRING")==0 || strcmp("HSG_CODE_SYM",token)==0 ||
	strcmp(token,"FORMAT")==0) {
      Tcl_AppendResult(CurrentInterp,
             "\nPossible cause:  code type where numeric required",NULL);
    }

    /*
     * Type mismatch
     */
    else if (strcmp(token,"FLOAT")==0 || strcmp("INT",token)==0 ||
	     strcmp(token,"HSG_NUM_SYM")==0) {
      Tcl_AppendResult(CurrentInterp,
           "\nPossible cause:  numeric type where code required",NULL);
    }
  }
  else
#endif

    Tcl_AppendResult(CurrentInterp,s,NULL);
}



/*
 *------------------------------------------------------------
 *
 * analyze_buff_init --
 *
 *      Called when initializing the parser.  Sets up private
 *      varaibles to hold the string to be parsed.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *     None.
 *
 *------------------------------------------------------------
 */
static void 
analyze_buff_init(string) 
     char CONST *string;
{
  if (sdbTextBuff) ckfree(sdbTextBuff);
  if (!(sdbTextBuff = (char*)ckalloc(strlen(string)+1))) {
    fprintf(stderr,"Out of memory");
    exit(1);
  }
  sdbTextPos = 0;
  strcpy(sdbTextBuff,string);
  sdbTextLen = strlen(sdbTextBuff);
}

/*------------------------------------------------------------
 *
 * Lexical analyzer rules follow
 *
 *----------------------------------------------------------*/

%}

token [a-zA-Z]+[a-zA-Z0-9_]*

%%


%{ 
/********** Special syntactic forms tokens **********/ 
%}
"IF"       { return IF; }
"format"       { return FORMAT; }
"interpolate" { return INTERPOLATE; }
"IN"       { return IN; }
"random"   { return RANDOM; }

%{ 
/********** Decimal numbers **********/ 
%}
([0-9]+\.[0-9]*)|([0-9]*\.[0-9]+) { 
  yylval.value = atof(yytext); 
  return FLOAT; 
}

%{ 
/********** Integer numbers **********/ 
%}
[0-9]+        { 
  yylval.integer = atoi(yytext); 
  return INT; 
}

%{ 
/********** Whitespace **********/ 
%}
[ \t\r\n]         {  } ;

%{ 
/********** Quoted text **********/ 
%}
\"[^"]* {  
     if (yytext[yyleng-1] == '\\') {
       yyleng--;
       yymore();
     }
     else {
       int new;
       char c;

       /* remove the trailing " " from the input stream */
       c=input();
       yylval.string = (Tcl_CreateHashEntry(htbl,yytext+1,&new)->key.string);
       return STRING;
    }
                         } ;

%{ 
/********** Literal lists **********/ 
%}
\"({token}\/)+{token}\" {  int new;
     yylval.string = (Tcl_CreateHashEntry(htbl,yytext,&new)->key.string);
                           return MULTI_STRING;
                         } ;

%{ 
/********** Tokens **********/ 
%}
\{.*\} |
{token}   { 

int i, new;
SdbVariable *varPtr;
char *name;

    if (yytext[0]=='{' && yytext[yyleng]=='}') {
      name=yytext+1;
      yytext[yyleng-1]=0;
    }
    else
      name=yytext;

       /* check for reserved symbols */ 
       if ((i = SdbNameCheck(name)) != UNKNOWN) {
         return i;
       }
     yylval.string = (Tcl_CreateHashEntry(htbl,name,&new)->key.string);
       /* check for HSG File variables */
       if ((varPtr=SdbFound(VarPtr,name))) {
	   if (varPtr->flag & SDB_NUMBER)
	     return HSG_NUM_SYM;
	   else
	     return HSG_CODE_SYM;
	 }
return UNKNOWN;
}

%{ 
/********** Special single character elements **********/ 
%}
\%       { return MOD; }
\^       { return POW; }
\-       { return MINUS; }
\+       { return PLUS; }
\*       { return MUL; }
\/       { return DIV; }
==       { return EQ; }
\<        { return LT; }
\<=       { return LE; }
\>        { return GT; }
\>=       { return GE; }
\!=       { return NE; }
\&\&       { return AND; }
\|\|       { return OR; }
\!         { return NOT; }

\;      { return EOS; } 
\:      { return CONNECTOR; }
\002    { return EOI; }

.        { return yytext[0]; }

%{ 
/********** End of lexical rules **********/ 
%}
%%
;

/* 
 * User code section. This area contains a few helper
 * functions for the lexical analyzer.  In particular,
 * built in functions are defined here.
 */
static double SdbRound _ANSI_ARGS_((double));
static double SdbInt _ANSI_ARGS_((double));

/*
 * Functions that take a single list argument and
 * return a double.
 */
static struct {
  char* name;
  double (*fnctptr)();
} hsgfnct[] = {
    {0         , 0  } };

/*
 * Functions that take a single string argument and
 * return a double.
 */
static struct {
    char* name;
    double (*fnctptr) _ANSI_ARGS_((CONST char*));
} c_fnct[] = {
    {"atof" , atof },
    {0      , 0     } };

/*
 * Functions that take a single double argument and
 * return a double.
 */
static struct {
    char* name;
    double (*fnctptr) _ANSI_ARGS_((double));
} s_fnct[] = {
    {"abs"  , fabs },
    {"acos" , acos },
    {"asin" , asin }, 
    {"atan" , atan },
    {"ceil",  ceil },
    {"cos"  , cos  },
    {"cosh" , cosh },
    {"exp"  , exp  },
    {"floor", floor},
    {"int"  , SdbInt },
    {"log"  , log  },
    {"log10", log10},
    {"round", SdbRound},
    {"sin"  , sin  },
    {"sinh" , sinh },
    {"sqrt" , sqrt },
    {"tan"  , tan  }, 
    {"tanh" , tanh },
    {0      , 0    } };

/*
 * Functions that take two double arguments and
 * return a double.
 */
static struct {
    char* name;
    double (*fnctptr) _ANSI_ARGS_((double , double));
} d_fnct[] = {
    {"atan2", atan2},
    {"hypot", hypot},
    {"fmod"  , fmod},
    {"pow"  , pow } ,
    {0      , 0    }  };



/*
 *------------------------------------------------------------
 *
 * SdbNameCheck --
 *
 *      Called by the lexer.  Given a string it searchs the function
 *      definitions for a match.  If found, it returns the token type.
 *
 * Results:
 *      Token Type (types are defined in the parser input file).
 *
 * Side effects:
 *      Sets the lexer return value to the address of the function.
 *
 *------------------------------------------------------------
 */
int 
SdbNameCheck(string) 
     char *string;
{

int i;

       /* check for builtin HSG functions */
       for (i=0; hsgfnct[i].name; i++) {
	 if (strcmp(hsgfnct[i].name, string) == 0) {
	   Sdblval.fnctptr = hsgfnct[i].fnctptr; 
	   return HSG_FNCT;
	 }
       }

       /* Check for builtin simplex math functions */
       for (i=0; s_fnct[i].name; i++) {
	 if (strcmp(s_fnct[i].name, string) == 0) {
	   Sdblval.fnctptr = s_fnct[i].fnctptr; 
	   return S_FNCT;
	 }
       }
       /* Check for builtin simplex character functions */
       for (i=0; c_fnct[i].name; i++) {
	 if (strcmp(c_fnct[i].name, string) == 0) {
	   Sdblval.fnctptr = c_fnct[i].fnctptr; 
	   return C_FNCT;
	 }
       }
       /* Check for builtin duplex math functions */
       for (i=0; d_fnct[i].name; i++) {
	 if (strcmp(d_fnct[i].name, string) == 0) {
	   Sdblval.fnctptr = d_fnct[i].fnctptr; 
	   return D_FNCT;
	 }
       }

return UNKNOWN;
}


/*
 *------------------------------------------------------------
 *
 * SdbReservedName --
 *
 *      Called by the lexer.  
 *      definitions for a match.  If found, it returns the token type.
 *
 * Results:
 *      Token Type (types are defined in the parser input file).
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
int 
SdbReservedName(string) 
     char *string;
{
  int i;
  if ((i=SdbNameCheck(string)) == UNKNOWN)
    return 0;
  else
    return 1;
}
  

/*
 *------------------------------------------------------------
 *
 * SdbNameCheck --
 *
 *      Called by SdbPrettyPrint (and possibly others).  Given
 *      a function pointer searches the defintions for a match.
 *      *
 * Results:
 *      If found, returns the subroutine name.  Otherwise
 *      returns " error! ".
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
char*
SdbWhatFnct(fnctptr) 
     double (*fnctptr)();
{

  int i;

       /* check for builtin HSG functions */
       for (i=0; hsgfnct[i].name; i++) {
	 if (hsgfnct[i].fnctptr == fnctptr) 
	   return(hsgfnct[i].name);
       }

       /* Check for builtin simplex math functions */
       for (i=0; s_fnct[i].name; i++) {
	 if (s_fnct[i].fnctptr == fnctptr) 
	   return(s_fnct[i].name);
       }

       /* Check for builtin duplex math functions */
       for (i=0; d_fnct[i].name; i++) {
	 if (d_fnct[i].fnctptr == fnctptr) 
	   return(d_fnct[i].name);
       }

       /* Check for character functions */
       for (i=0; c_fnct[i].name; i++) {
	 if (c_fnct[i].fnctptr == fnctptr) 
	   return(c_fnct[i].name);
       }

return(" error! ");
}



/*
 *------------------------------------------------------------
 *
 * SdbRound, SdbInt --
 *
 *      Emulate the TCL math functions of the same name.
 *      These functions are slightly different, they
 *      do not convert to integer and do not trap errors.
 *      Error handling is left to the OS.
 *
 * Results:
 *      SdbInt = Integer value produced by truncating towards 0
 *      SdbRound = Integer value produced by rounding x
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static double
SdbRound (x)
     double x;
{
  return(floor(x+0.5));
}

static double
SdbInt (x)
     double x;
{
  if (x>=0)
    return(floor(x));
  else
    return(ceil(x));
}
